from typing import Optional, List

from pydantic import BaseModel, UUID4

from nidhoggr.core.user import UserProperty
from nidhoggr.utils.transformer import YggdrasilRequestTransformer, JSONResponseTransformer, LegacyRequestTransformer


class JoinRequest(BaseModel, YggdrasilRequestTransformer):
    accessToken: UUID4
    selectedProfile: UUID4
    serverId: str

    class Config:
        allow_mutation = False


class HasJoinedRequestBase(BaseModel):
    username: str
    serverId: str
    ip: Optional[str]


class HasJoinedRequest(HasJoinedRequestBase, YggdrasilRequestTransformer):

    class Config:
        allow_mutation = False


class HasJoinedRequestLegacy(HasJoinedRequestBase, LegacyRequestTransformer):

    class Config:
        allow_mutation = False


class JoinedResponse(BaseModel, JSONResponseTransformer):
    id: UUID4
    name: str
    properties: List[UserProperty]

    class Config:
        allow_mutation = False


class ProfileRequestBase(BaseModel):
    id: UUID4
    unsigned: bool = False

    class Config:
        allow_mutation = False


class ProfileRequest(ProfileRequestBase, YggdrasilRequestTransformer):

    class Config:
        allow_mutation = False


class ProfileRequestLegacy(ProfileRequestBase, LegacyRequestTransformer):

    class Config:
        allow_mutation = False



from uuid import UUID

from nidhoggr.errors.auth import InvalidAccessToken, InvalidClientToken
from nidhoggr.errors.common import BadPayload
from ..conftest import cast


def test_empty_credentials(user, validate):
    assert cast[BadPayload](validate({}))
    assert cast[BadPayload](validate({"clientToken": user.client}))


def test_broken_access_token(user, validate):
    response = validate({
        "accessToken": UUID('04ee4899-880d-4a36-b6de-d62cd581328d'),
        "clientToken": user.client
    })
    assert response.status_code == 403
    assert cast[InvalidAccessToken](response)


def test_invalid_client_token(user, validate):
    response = validate({
        "accessToken": user.access,
        "clientToken": UUID('30e4b8ed-e8d4-4e6f-b651-e4d700a4606c')
    })
    assert response.status_code == 403
    assert cast[InvalidClientToken](response)


def test_validate_full(user, validate):
    response = validate({
        "accessToken": user.access,
        "clientToken": user.client
    })
    assert response.status_code == 204
    assert response.data == b""


def test_validate_simple(user, validate):
    response = validate({
        "accessToken": user.access,
    })
    assert response.status_code == 204
    assert response.data == b""



import json
from functools import wraps
from itertools import product
from pathlib import Path
from random import choice
from string import ascii_lowercase
from typing import Dict, Any, Union, Callable, TypeVar
from uuid import UUID, uuid4

import pytest
from flask import url_for, Response, Flask
from flask.testing import FlaskClient
from nidhoggr.core.response import StatusResponse, ErrorResponse, TextureStatusResponse

from nidhoggr.application import create_app
from nidhoggr.core.user import User
from nidhoggr.core.config import BLConfig
from nidhoggr.core.repository import BaseTextureRepo, BaseUserRepo
from nidhoggr.errors.common import YggdrasilError
from nidhoggr.models.auth import AuthenticationResponse, RefreshResponse
from nidhoggr.models.session import JoinedResponse
from nidhoggr.core.texture import TextureRequest, TextureUploadRequest, ComplexTextureResponse
from nidhoggr.utils.crypto import generate_keypair


class TestUser(User):
    # NOTE: Still needed for testing purposes
    password: str


class UUIDEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, UUID):
            # if the obj is uuid, we simply return the value of uuid
            return obj.hex
        return json.JSONEncoder.default(self, obj)


USERS = list(map(
    TestUser.parse_obj,
    json.loads((Path(__file__).parent / "data/users.json").read_text())
))

EndpointCallable = Callable[[Dict[str, Any]], Response]
TEST_KEY_PAIR = generate_keypair()

YggdrasilResponse = Union[
    AuthenticationResponse,
    RefreshResponse,
    JoinedResponse,
]


def accessor(func: Callable):
    @wraps(func)
    def wrapper(client: FlaskClient, data: Dict[str, Any], query_data: Dict[str, Any] = None, swap: bool = False):
        return client.post(
            url_for(func.__name__),
            data=json.dumps(data, cls=UUIDEncoder),
            content_type='application/json'
        )

    return wrapper


def accessor_legacy(func: Callable):
    @wraps(func)
    def wrapper(client: FlaskClient, data: Dict[str, Any], query_data: Dict[str, Any], swap: bool = False):
        if swap:
            return client.get(
                url_for(func.__name__),
                query_string=data,
            )
        else:
            return client.get(
                url_for(func.__name__, **data),
                query_string=query_data,
            )

    return wrapper


T = TypeVar("T")


class cast:
    def __getitem__(self, other) -> Callable[[T], Union[YggdrasilError, YggdrasilResponse]]:
        if isinstance(other, YggdrasilError):
            return lambda res: self[YggdrasilError](res) == other
        elif issubclass(other, YggdrasilError):
            return lambda res: other(status=res.status_code, **res.json)
        elif other in YggdrasilResponse.__args__:
            return lambda res: other(**res.json)
        else:
            return lambda res: None


cast = cast()


class TestUserRepo(BaseUserRepo):
    __users: Dict[int, TestUser]

    def __init__(self, *, users):
        self.__users = {user.uuid: user.copy() for user in users}

    def get_user(self, **kw: Dict[str, str]) -> Union[ErrorResponse, User]:
        res = [
            u
            for u
            in self.__users.values()
            if any(getattr(u, k) == v for k, v in kw.items())
        ]
        return (res or [BaseUserRepo.EMPTY_USER])[0]

    def save_user(self, *, user: TestUser) -> Union[ErrorResponse, User]:
        self.__users[user.uuid] = user
        return user

    def check_password(self, *, clean: str, uuid: str) -> Union[ErrorResponse, StatusResponse]:
        user = self.get_user(uuid=uuid)
        status = user and user.password == clean
        return StatusResponse(status=status)


# FIXME
class TestTextureRepo(BaseTextureRepo):
    def get(self, *, request: TextureRequest) -> Union[ErrorResponse, ComplexTextureResponse]:
        return ComplexTextureResponse(
            timestamp=0,
            profileId=uuid4(),
            profileName="",
            textures={}
        )

    def upload(self, *, request: TextureUploadRequest) -> Union[ErrorResponse, TextureStatusResponse]:
        return TextureStatusResponse(message="OK")

    def clear(self, *, request: TextureRequest) -> Union[ErrorResponse, TextureStatusResponse]:
        TextureStatusResponse(message="OK")


@pytest.fixture(params=list(product((False, True), (False, True), (False, True))))
def config(request) -> BLConfig:
    strict, simple_login, legacy_compat, *_ = request.param
    return BLConfig(
        key_pair=TEST_KEY_PAIR,
        strict=strict,
        simple_login=simple_login,
        legacy_compat=legacy_compat,
    )


@pytest.fixture
def users() -> TestUserRepo:
    return TestUserRepo(users=USERS)


@pytest.fixture
def textures() -> TestTextureRepo:
    return TestTextureRepo()


@pytest.fixture(autouse=True)
def skip_bl(request, config):
    if request.node.get_closest_marker("skip_bl_on"):
        marker = request.node.get_closest_marker("skip_bl_on")
        if any(
            marker.kwargs.get(name, ...) == value
            for name, value
            in config._asdict().items()
        ):
            pytest.skip(f"Skipped on {config}")


@pytest.fixture
def user(users) -> User:
    return users.get_user(uuid=UUID("3388364d-4e96-4567-aaf9-3d113cdef2af"))


@pytest.fixture
def old_user(users) -> User:
    return users.get_user(uuid=UUID("0b6b2656-73fb-41ce-b116-c3c4cc5753c4"))


@pytest.fixture
def new_user(users) -> User:
    return users.get_user(uuid=UUID("3ad95d5d-7986-4464-837f-dd6dd7444ed5"))


@pytest.fixture
def app(
    users,
    config: BLConfig,
    textures: BaseTextureRepo
) -> Flask:
    return create_app(
        users=users,
        config=config,
        textures=textures,
    )


def generate_server_id() -> str:
    return ''.join(choice(ascii_lowercase) for i in range(40))



from datetime import datetime
from typing import Union

from nidhoggr.core.config import BLConfig
from nidhoggr.core.repository import BaseTextureRepo, BaseUserRepo
from nidhoggr.core.texture import TextureType, TextureRequest, ComplexTextureResponse
from nidhoggr.core.user import User

from nidhoggr.errors.auth import InvalidProfile
from nidhoggr.errors.common import YggdrasilError
from nidhoggr.errors.session import InvalidServer
from nidhoggr.models.auth import EmptyResponse, NoContent
from nidhoggr.models.session import JoinRequest, HasJoinedRequest, JoinedResponse, ProfileRequest
from nidhoggr.utils.crypto import sign_property
from nidhoggr.utils.decorator import typed
from nidhoggr.utils.repository import handle_status


@typed
def join(req: JoinRequest, users: BaseUserRepo) -> Union[EmptyResponse, YggdrasilError]:
    user: User = handle_status(users.get_user(access=req.accessToken))

    if user.synthetic:
        return InvalidProfile

    if req.selectedProfile != user.uuid:
        return InvalidProfile

    user = user.copy(update=dict(server=req.serverId))
    handle_status(users.save_user(user=user))
    return NoContent


@typed
def has_joined(req: HasJoinedRequest, users: BaseUserRepo) -> Union[JoinedResponse, YggdrasilError]:
    user: User = handle_status(users.get_user(login=req.username))

    if user.synthetic:
        return InvalidProfile

    if req.serverId != user.server:
        return InvalidServer

    return JoinedResponse(
        id=user.uuid,
        name=user.login,
        properties=user.properties
    )


@typed
def profile(
    req: ProfileRequest,
    users: BaseUserRepo,
    config: BLConfig,
    textures: BaseTextureRepo
) -> Union[JoinedResponse, YggdrasilError]:
    user: User = handle_status(users.get_user(uuid=req.id))

    if user.synthetic:
        return InvalidProfile

    request = TextureRequest(uuid=req.id, texture_types=list(TextureType))
    texture_response = handle_status(textures.get(request=request))

    texture_prop = ComplexTextureResponse(
        timestamp=int(datetime.now().timestamp() * 1000),
        profileId=user.uuid,
        profileName=user.login,
        textures=texture_response.textures
    ).pack()

    raw_props = [texture_prop, *user.properties]

    if req.unsigned:
        properties = raw_props
    else:
        properties = [
            sign_property(private_key=config.key_pair.private, prop=prop)
            for prop
            in raw_props
        ]

    return JoinedResponse(
        id=user.uuid,
        name=user.login,
        properties=properties
    )


__all__ = {"join", "has_joined", "profile"}



from typing import Union

from nidhoggr.core.config import BLConfig
from nidhoggr.core.repository import BaseTextureRepo, BaseUserRepo
from nidhoggr.errors.common import YggdrasilError
from nidhoggr.models.session import HasJoinedRequestLegacy, ProfileRequestLegacy, JoinedResponse
from nidhoggr.utils.decorator import typed
from nidhoggr.views import session


# noinspection PyPep8Naming
@typed
def hasJoined(req: HasJoinedRequestLegacy, users: BaseUserRepo) -> Union[JoinedResponse, YggdrasilError]:
    return session.has_joined.__wrapped__(req, users)


@typed
def profile(
    req: ProfileRequestLegacy,
    users: BaseUserRepo,
    config: BLConfig,
    textures: BaseTextureRepo
) -> Union[JoinedResponse, YggdrasilError]:
    return session.profile.__wrapped__(req, users, config, textures)


__all__ = {"hasJoined", "profile"}



from typing import Optional
from uuid import uuid4

import pytest

from nidhoggr.errors.auth import InvalidProfile
from nidhoggr.errors.common import BadPayload
from nidhoggr.errors.session import InvalidServer
from nidhoggr.models.session import JoinedResponse
from ..conftest import cast, generate_server_id


optional_ip = pytest.mark.parametrize(
    ["ip"],
    [("127.0.0.1",), (None,)],
    ids=["exists", "absent"]
)


def test_empty_credentials(user, has_joined):
    assert cast[BadPayload](has_joined({}, {}))
    assert cast[BadPayload](has_joined({"username": user.login}, {}, True))
    assert cast[BadPayload](has_joined({"ip": "127.0.0.1"}, {}, True))
    assert cast[BadPayload](has_joined({"serverId": uuid4()}, {}, True))
    assert cast[BadPayload](has_joined({"username": user.login, "ip": "127.0.0.1"}, {}, True))
    assert cast[BadPayload](has_joined({"ip": "127.0.0.1", "serverId": uuid4()}, {}, True))


@optional_ip
def test_invalid_server(user, join, has_joined, ip: Optional[str]):
    join({
        "accessToken": user.access,
        "selectedProfile": user.uuid,
        "serverId": generate_server_id()
    })
    has_joined_response = has_joined({
        "username": user.login,
        "serverId": "anything",
        "ip": ip,
    }, {}, True)

    assert cast[InvalidServer](has_joined_response)


@optional_ip
def test_nonexistent_user(has_joined, ip: Optional[str]):
    response = has_joined({
        "username": "anything",
        "serverId": "anything",
        "ip": ip,
    }, {}, True)

    assert cast[InvalidProfile](response)


@optional_ip
def test_has_joined(user, join, has_joined, ip: Optional[str]):
    server_id = generate_server_id()
    join({
        "accessToken": user.access,
        "selectedProfile": user.uuid,
        "serverId": server_id
    })
    has_joined_response = has_joined({
        "username": user.login,
        "serverId": server_id,
        "ip": "127.0.0.2",
    }, {}, True)
    result = cast[JoinedResponse](has_joined_response)
    assert result is not None
    assert result.id == user.uuid
    assert result.name == user.login



from typing import Union, TypeVar

from nidhoggr.core.response import ErrorResponse
from werkzeug.exceptions import InternalServerError

T = TypeVar("T")


def handle_status(repository_response: Union[ErrorResponse, T]) -> T:
    if isinstance(repository_response, ErrorResponse):
        raise InternalServerError(response=repository_response.reason, description=repository_response.exception)
    return repository_response



from typing import Union
from uuid import uuid4

from nidhoggr.core.config import BLConfig
from nidhoggr.core.repository import BaseUserRepo
from nidhoggr.core.user import User

from nidhoggr.errors.auth import MigrationDone, InvalidCredentials, InvalidAccessToken, InvalidClientToken
from nidhoggr.errors.common import YggdrasilError, BadRequest
from nidhoggr.models import auth
from nidhoggr.utils.decorator import typed
from nidhoggr.utils.repository import handle_status


@typed
def authenticate(
    req: auth.AuthenticationRequest,
    users: BaseUserRepo,
    config: BLConfig
) -> Union[auth.AuthenticationResponse, YggdrasilError]:
    user: User = handle_status(users.get_user(email=req.username, login=req.username))

    if user.synthetic:
        return InvalidCredentials

    password_check = handle_status(users.check_password(clean=req.password, uuid=user.uuid))
    if not password_check.status:
        return InvalidCredentials

    if not config.simple_login and req.username == user.login:
        return MigrationDone

    if (config.strict and (user.client is None)) or (not config.strict):
        client_token = req.clientToken or uuid4()
        user = user.copy(update=dict(client=client_token))
    elif config.strict and req.clientToken is not None and user.client != req.clientToken:
        return InvalidClientToken

    user = user.copy(update=dict(access=uuid4()))

    if req.agent is not None:
        profile = auth.Profile(id=user.uuid, name=user.login)
        selected_profile, available_profiles = profile, [profile]
    else:
        selected_profile = available_profiles = None

    if req.requestUser:
        user_instance = auth.UserInstance(id=user.uuid, properties=user.properties)
    else:
        user_instance = None

    handle_status(users.save_user(user=user))

    return auth.AuthenticationResponse(
        accessToken=user.access,
        clientToken=user.client,
        availableProfiles=available_profiles,
        selectedProfile=selected_profile,
        user=user_instance,
    )


@typed
def refresh(req: auth.RefreshRequest, users: BaseUserRepo) -> Union[auth.RefreshResponse, YggdrasilError]:
    user: User = handle_status(users.get_user(client=req.clientToken))

    if user.synthetic:
        return BadRequest

    user = user.copy(update=dict(access=uuid4()))
    selected_profile = auth.Profile(id=user.uuid, name=user.login)

    if req.requestUser:
        user_instance = auth.UserInstance(id=user.uuid, properties=user.properties)
    else:
        user_instance = None

    handle_status(users.save_user(user=user))

    return auth.RefreshResponse(
        accessToken=user.access,
        clientToken=user.client,
        selectedProfile=selected_profile,
        user=user_instance,
    )


@typed
def validate(req: auth.ValidateRequest, users: BaseUserRepo) -> Union[auth.EmptyResponse, YggdrasilError]:
    user: User = handle_status(users.get_user(access=req.accessToken))

    if user.synthetic:
        return InvalidAccessToken

    if req.clientToken is not None and req.clientToken != user.client:
        return InvalidClientToken

    return auth.NoContent


@typed
def invalidate(req: auth.InvalidateRequest, users: BaseUserRepo) -> Union[auth.EmptyResponse, YggdrasilError]:
    user: User = handle_status(users.get_user(access=req.accessToken))

    if user.synthetic:
        return InvalidAccessToken

    if req.clientToken != user.client:
        return InvalidClientToken

    user = user.copy(update=dict(access=None))

    handle_status(users.save_user(user=user))

    return auth.NoContent


@typed
def signout(req: auth.SignOutRequest, users: BaseUserRepo) -> Union[auth.EmptyResponse, YggdrasilError]:
    user: User = handle_status(users.get_user(email=req.username))

    if user.synthetic:
        return InvalidCredentials

    password_check = handle_status(users.check_password(clean=req.password, uuid=user.uuid))
    if not password_check.status:
        return InvalidCredentials

    user = user.copy(update=dict(access=None))

    handle_status(users.save_user(user=user))

    return auth.NoContent


__all__ = {"authenticate", "refresh", "validate", "invalidate", "signout"}



from base64 import b64encode, b64decode
from uuid import uuid4

from M2Crypto import EVP, RSA, BIO

from nidhoggr.core.crypto import KeyPair
from nidhoggr.core.user import UserProperty


def generate_keypair(*, bits: int = 4096) -> KeyPair:
    """
    Generate an RSA keypair with an exponent of 65537 in PEM format

    :param: bits The key length in bits
    :return: Key pair with private and public keys
    """
    key = RSA.gen_key(bits, 65537)

    with BIO.MemoryBuffer() as buffer:
        key.save_key_bio(buffer, cipher=None)
        private_key = buffer.getvalue()

    with BIO.MemoryBuffer() as buffer:
        key.save_pub_key_bio(buffer)
        public_key = buffer.getvalue()

    del key
    return KeyPair(private=private_key, public=public_key)


def sign_property(*, private_key: bytes, prop: UserProperty) -> UserProperty:
    key = EVP.load_key_string(private_key)
    key.reset_context(md="sha1")
    key.sign_init()
    key.sign_update(prop.value.encode("ascii"))
    signature = b64encode(key.sign_final()).decode("ascii")
    return prop.copy(update=dict(signature=signature))


def verify_property(*, public_key: bytes, prop: UserProperty) -> bool:
    if prop.signature is None:
        return False
    with BIO.MemoryBuffer(public_key) as buffer:
        key = EVP.load_key_bio_pubkey(buffer)
    key.reset_context(md="sha1")
    key.verify_init()
    key.verify_update(prop.value.encode("ascii"))
    result = (key.verify_final(b64decode(prop.signature)) == 1)
    del key
    return result


def generate_uuid() -> str:
    """Generate random UUID token like Java's UUID.toString()"""
    return uuid4().hex



from uuid import UUID

import pytest

from nidhoggr.core.user import User, UserProperty
from nidhoggr.core.config import BLConfig
from nidhoggr.errors.auth import InvalidProfile
from nidhoggr.errors.common import BadPayload
from nidhoggr.models.session import JoinedResponse
from nidhoggr.utils.crypto import verify_property
from ..conftest import cast

with_sign = pytest.mark.parametrize(
    ["unsigned"],
    [(True,), (False,)],
    ids=["unsigned", "signed"]
)


@pytest.mark.skip_bl_on(legacy_compat=True)
def test_empty_credentials(profile):
    assert cast[BadPayload](profile({"id": ""}, {}))


@with_sign
def test_nonexistent_user(unsigned: bool, profile):
    assert cast[InvalidProfile](profile({"id": UUID('facdbba0-f4b5-47e8-be9f-f7124e12ec22')}, {"unsigned": unsigned}))


@with_sign
def test_profile_simple(user: User, unsigned: bool, profile):
    response = profile({"id": user.uuid}, {"unsigned": unsigned})
    result = cast[JoinedResponse](response)
    assert result is not None
    assert result.id == user.uuid
    assert result.name == user.login


@with_sign
def test_properties(config: BLConfig, new_user: User, unsigned: bool, profile):
    response = profile({"id": new_user.uuid}, {"unsigned": unsigned})
    result = cast[JoinedResponse](response)
    assert result is not None
    if not unsigned:
        assert all(verify_property(public_key=config.key_pair.public, prop=p) for p in result.properties)
    assert set(p.unsigned for p in result.properties) >= set(new_user.properties)


@with_sign
def test_texture_exists(user: User, unsigned: bool, profile):
    response = profile({"id": user.uuid}, {"unsigned": unsigned})
    result = cast[JoinedResponse](response)
    assert result is not None
    texture_prop = ([p for p in result.properties if p.name == "textures"] or [None])[0]
    assert isinstance(texture_prop, UserProperty)


@with_sign
def test_texture_attached(user: User, unsigned: bool, profile):
    response = profile({"id": user.uuid}, {"unsigned": unsigned})
    result = cast[JoinedResponse](response)
    assert result is not None
    diff = set(p.unsigned for p in result.properties) - set(user.properties)
    assert len(diff) == 1
    assert diff.pop().name == "textures"



from uuid import uuid4, UUID

from nidhoggr.core.repository import BaseUserRepo
from nidhoggr.errors.auth import InvalidProfile
from nidhoggr.errors.common import BadPayload
from ..conftest import cast, generate_server_id


def test_empty_credentials(user, join):
    assert cast[BadPayload](join({}))
    assert cast[BadPayload](join({"accessToken": user.access}))
    assert cast[BadPayload](join({"selectedProfile": user.uuid}))
    assert cast[BadPayload](join({"serverId": uuid4()}))
    assert cast[BadPayload](join({"accessToken": user.access, "selectedProfile": user.uuid}))
    assert cast[BadPayload](join({"accessToken": user.access, "serverId": uuid4()}))
    assert cast[BadPayload](join({"selectedProfile": user.access, "serverId": uuid4()}))


def test_invalid_profile(user, join):
    response = join({
        "accessToken": user.access,
        "selectedProfile": UUID('df7e45e0-ebe4-40ad-a744-16ee1a04051f'),
        "serverId": uuid4()
    })
    assert cast[InvalidProfile](response)


def test_nonexistent_user(user, join):
    server_id = generate_server_id()
    response = join({
        "accessToken": UUID('0f2d2d09-5a4a-4e88-b774-bee307f6f107'),
        "selectedProfile": user.uuid,
        "serverId": server_id
    })

    assert cast[InvalidProfile](response)


def test_join(users: BaseUserRepo, user, join):
    server_id = generate_server_id()
    response = join({
        "accessToken": user.access,
        "selectedProfile": user.uuid,
        "serverId": server_id
    })
    assert response.status_code == 204
    assert response.data == b""
    fresh = users.get_user(uuid=user.uuid)
    assert fresh.server == server_id
